/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Chroma widget set
 * (C) 2005 M. Derezynski <internalerror@gmail.com>
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>

#include "main.h"
#include "bmpx/util.h"
#include "ui_util.h"

#include "chroma_list.h"
#include "chroma_vscrollbar.h"

#include "marshalers.h"
#include "g_object_get.h"


#define CHROMA_LIST_SELECTION_ALPHA 0.4

G_DEFINE_TYPE(ChromaList, chroma_list, G_TYPE_OBJECT)

G_LOCK_DEFINE(LIST_WINDOW);

typedef enum {
    CHROMA_LIST_SIGNAL_EXPOSE_REQUEST,
    CHROMA_LIST_SIGNAL_CHANGED,
    CHROMA_LIST_SIGNAL_ROW_ACTIVATED,
    CHROMA_LIST_SIGNAL_CURSOR_POSITION,

    CHROMA_LIST_N_SIGNALS
} ChromaListSignals;
static guint chroma_list_signals[CHROMA_LIST_N_SIGNALS] = {0,0,0,0};

typedef enum {
    C_LIST_PROP_X = 1,
    C_LIST_PROP_Y,
    C_LIST_PROP_W,
    C_LIST_PROP_H,
    C_LIST_PROP_LIST_OFFSET,
    C_LIST_PROP_N_ITEMS,
    C_LIST_PROP_ROW_HEIGHT,

    C_LIST_N_PROPERTIES
} ChromaListProperties;

struct _ChromaListPrivate {
    gboolean dispose_has_run;

    BmpListType *list;

    BmpFontInformation *font_information;
    gint row_height;
    gint list_offset;

    gboolean active;
    gint grabbed_position;

    /* Parameters */
    GtkSelectionMode selection_mode;
    ChromaListDropMode drop_mode;
    ChromaListAlignMode align;
    gint columns;
    gboolean use_numbers;
    gboolean change_on_select;

    gboolean drag;
    gint drag_row;

    gboolean changed;

    /* Drawing stuff */
    GtkAllocation allocation;
    GdkPixmap *window;
    GtkWidget *parent;

    GtkCellRenderer *cellrenderer;
};

typedef struct {
    ChromaList *self;
    gchar *name;
    gint idx;
} SelectNameTuple;

/* Rendering Fcts Prototypes */
static void
process_cells (ChromaList *self);

static void
recreate_window(ChromaList * self);

static void
mark_visible_dirty (ChromaList *self);

static void
mark_visible_dirty_nonrendered (ChromaList *self);

static void
blank_window (ChromaList *self);

static void
scroll_up (ChromaList *self);

static void
scroll_down (ChromaList *self);

static void
page_up (ChromaList *self);

static void
page_down (ChromaList *self);

void
chroma_list_emit_expose_request(ChromaList * self)
{
    g_signal_emit(G_OBJECT(self),
                  chroma_list_signals[CHROMA_LIST_SIGNAL_EXPOSE_REQUEST], 0,
                  g_object_get_int(self, "pos_x"), g_object_get_int(self,
                                                                    "pos_y"),
                  g_object_get_int(self, "width"), g_object_get_int(self,
                                                                    "height"));
}

void
chroma_list_emit_cursor_position (ChromaList * self)
{
    g_signal_emit(G_OBJECT(self),
                  chroma_list_signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION], 0, self->private->list->last_selected);
}

void
chroma_list_set_selection_mode(ChromaList * self,
                               GtkSelectionMode selection_mode)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    if ((selection_mode != GTK_SELECTION_MULTIPLE) &&
        (selection_mode != GTK_SELECTION_SINGLE))
        return;

    self->private->selection_mode = selection_mode;

}

void
chroma_list_set_columns(ChromaList * self, gint columns)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->columns = columns;

    chroma_list_emit_expose_request(CHROMA_LIST(self));
}

void
chroma_list_set_align(ChromaList * self, ChromaListAlignMode align)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->align = align;

    chroma_list_emit_expose_request(CHROMA_LIST(self));
}

static void
render_row (ChromaList *self, gint k, gint k_item, gint left_padding) {

    GdkColor *fg_normal, *fg_current, *bg_normal, *bg_selected;
    cairo_pattern_t *cr_pattern;
    cairo_t *cr;
    gint cell_width, cell_height, cell_x, cell_y;
    gchar *aux1, *column_text;
    gpointer list_item;
    gboolean selected = FALSE;
    gint n;
    double r, g, b, a, x, y, w, h;

    enum {
        RECT_BG,
        RECT_CELL,
        RECT_EXPOSE,

        N_RECTS
    };

    GdkRectangle *rect[N_RECTS];

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (!self->private->list->items)
	return;

    fg_normal =
	bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL);
    fg_current =
	bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT);
    bg_normal =
	bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL);
    bg_selected =
	bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_SELECTED);

    rect[RECT_BG] =
	g_new0(GdkRectangle, 1);
    rect[RECT_CELL] =
	g_new0(GdkRectangle, 1);
    rect[RECT_EXPOSE] =
	g_new0(GdkRectangle, 1);

    cr =
	gdk_cairo_create (self->private->window);

    /* Create selection cairo pattern */
    gdk_color_to_rgba(fg_current, &r, &g, &b, &a);
    cr_pattern = cairo_pattern_create_linear
        (0,
         self->private->row_height / 2,
         g_object_get_int(self, "width") - left_padding -
         (self->private->use_numbers ? self->private->font_information->
          font_width_approx_letters * 4 : 0), self->private->row_height / 2);

    cairo_pattern_add_color_stop_rgba(cr_pattern, 0.0, r, g, b, 0.4);
    cairo_pattern_add_color_stop_rgba(cr_pattern, 0.3, r, g, b, 0.28);
    cairo_pattern_add_color_stop_rgba(cr_pattern, 0.6, r, g, b, 0.20);
    cairo_pattern_add_color_stop_rgba(cr_pattern, 1, r, g, b, 0.15);
    cairo_pattern_set_filter(cr_pattern, CAIRO_FILTER_BILINEAR);

    list_item = g_list_nth(self->private->list->items, k_item)->data;
    if (!list_item)
      {
	cairo_pattern_destroy (cr_pattern);
	cairo_destroy (cr);
	return;
      }

    /* Clear row background */
    gdk_color_to_rgba(bg_normal, &r, &g, &b, &a);
    cell_width = g_object_get_int(self, "width");
    cell_height = self->private->row_height;
    cell_x = 0;
    cell_y = (k * self->private->row_height);
    cairo_rectangle(cr, cell_x, cell_y, cell_width, cell_height);
    cairo_set_source_rgba(cr, r, g, b, 1.0);
    cairo_fill(cr);

    selected = self->private->list->row_get_selected_func(list_item);

    if (self->private->list->on_display_func)
        self->private->list->on_display_func (list_item);

    g_object_set (G_OBJECT(self->private->cellrenderer), "font-desc", self->private->font_information->font_description, NULL);

    /* Render Idx Number */
    if (self->private->use_numbers) {

	g_object_set (self->private->cellrenderer, "background-gdk", bg_normal, NULL);

	aux1 = g_strdup_printf("%d", k_item+1);

	g_object_set (self->private->cellrenderer, "foreground-gdk", fg_normal, NULL);
	g_object_set (self->private->cellrenderer, "weight", PANGO_WEIGHT_NORMAL, NULL);

	if (self->private->list->current == k_item)
	  {
	    g_object_set (self->private->cellrenderer, "foreground-gdk", fg_current, NULL);
	    g_object_set (self->private->cellrenderer, "weight", PANGO_WEIGHT_NORMAL+100, NULL);
	  }

	cell_width = left_padding;
	cell_height = self->private->row_height;
	cell_x = 0;
	cell_y = (k * self->private->row_height);

	gtk_cell_renderer_set_fixed_size (GTK_CELL_RENDERER(self->private->cellrenderer),
						     cell_width,
						     cell_height);

	g_object_set (self->private->cellrenderer, "markup", aux1, NULL);
	g_object_set (self->private->cellrenderer, "xalign", 1.0, NULL);
	g_object_set (self->private->cellrenderer, "yalign", 0.5, NULL);
	g_object_set (self->private->cellrenderer, "xpad", 0, NULL);
	g_object_set (self->private->cellrenderer, "ypad", 0, NULL);
	g_object_set (self->private->cellrenderer, "ellipsize", PANGO_ELLIPSIZE_NONE, NULL);

	rect[RECT_BG]->x = cell_x;
	rect[RECT_BG]->y = cell_y;
	rect[RECT_BG]->width = cell_width;
	rect[RECT_BG]->height = cell_height;

	memcpy(rect[RECT_CELL], rect[RECT_BG], sizeof(GdkRectangle));
	memcpy(rect[RECT_EXPOSE], rect[RECT_BG], sizeof(GdkRectangle));

	gtk_cell_renderer_render (self->private->cellrenderer,
				     self->private->window,
				     self->private->parent,
				     rect[RECT_BG],
				     rect[RECT_CELL],
				     rect[RECT_EXPOSE], 0);

	g_free (aux1);
    }


    /* Render Text */
    cell_width =
	        g_object_get_int(self, "width") - left_padding -  (self->private->use_numbers ? 2*self->private->font_information->font_width_approx_digits : 0);
    cell_height =
		self->private->row_height;
    cell_x =
		left_padding + (self->private->use_numbers ? 2*self->private->font_information->font_width_approx_digits : 0);
    cell_y =
		(k * self->private->row_height);

    gtk_cell_renderer_set_fixed_size (self->private->cellrenderer,
					     cell_width,
					     cell_height);

    column_text = self->private->list->column_data_func (list_item, 0);
    g_object_set (self->private->cellrenderer, "text", column_text, NULL);
    g_free (column_text);

    if (self->private->align == CHROMA_LIST_ALIGN_LEFT)
		   g_object_set (self->private->cellrenderer, "xalign", 0.0, NULL);
    else
		   g_object_set (self->private->cellrenderer, "xalign", 0.5, NULL);

    g_object_set (self->private->cellrenderer, "yalign", 0.5, NULL);
    g_object_set (self->private->cellrenderer, "xpad", 0, NULL);
    g_object_set (self->private->cellrenderer, "ypad", 0, NULL);
    g_object_set (self->private->cellrenderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL);

    rect[RECT_BG]->x = cell_x;
    rect[RECT_BG]->y = cell_y;
    rect[RECT_BG]->width = cell_width;
    rect[RECT_BG]->height = cell_height;

    memcpy(rect[RECT_CELL], rect[RECT_BG], sizeof(GdkRectangle));
    memcpy(rect[RECT_EXPOSE], rect[RECT_BG], sizeof(GdkRectangle));

    if (!self->private->use_numbers)
      rect[RECT_CELL]->x += 4;

    if (self->private->columns == 2) rect[RECT_CELL]->width -= self->private->font_information->font_width_approx_digits * 7;

    g_object_set (self->private->cellrenderer, "background-set", FALSE, NULL);

    if (selected)
      {
		   	   g_object_set (self->private->cellrenderer, "foreground-gdk", fg_current, NULL);
		   	   g_object_set (self->private->cellrenderer, "weight", PANGO_WEIGHT_NORMAL + 50, NULL);

      }
    else
      {

		   	   g_object_set (self->private->cellrenderer, "foreground-gdk", fg_normal, NULL);
		   	   g_object_set (self->private->cellrenderer, "weight", PANGO_WEIGHT_NORMAL, NULL);
      }

    if (self->private->list->current == k_item)
      {
		   	   g_object_set (self->private->cellrenderer, "foreground-gdk", fg_current, NULL);
		   	   g_object_set (self->private->cellrenderer, "weight", PANGO_WEIGHT_NORMAL + 100, NULL);
      }

    x = rect[RECT_BG]->x;
    y = rect[RECT_BG]->y;
    w = rect[RECT_BG]->width;
    h = rect[RECT_BG]->height;

    gtk_cell_renderer_render (self->private->cellrenderer,
				     self->private->window,
				     self->private->parent,
				     rect[RECT_BG],
				     rect[RECT_CELL],
				     rect[RECT_EXPOSE], 0);


    /* Render Length */
    if (self->private->columns == 2) {

	   if (selected) {
		g_object_set (self->private->cellrenderer, "foreground-gdk", fg_current, NULL);
		g_object_set (self->private->cellrenderer, "weight", PANGO_WEIGHT_NORMAL + 50, NULL);
	   } else {
		g_object_set (self->private->cellrenderer, "foreground-gdk", fg_normal, NULL);
		g_object_set (self->private->cellrenderer, "weight", PANGO_WEIGHT_NORMAL, NULL);
	   }

	   if (self->private->list->current == k_item) {
		   g_object_set (self->private->cellrenderer, "foreground-gdk", fg_current, NULL);
		   g_object_set (self->private->cellrenderer, "weight", PANGO_WEIGHT_NORMAL + 100, NULL);
	   }

	   cell_width =
		self->private->font_information->font_width_approx_digits * 7;
           cell_height =
		self->private->row_height;
	   cell_x =
		g_object_get_int(self, "width") - self->private->font_information->font_width_approx_digits*7;
           cell_y =
		(k * self->private->row_height);

	   gtk_cell_renderer_set_fixed_size (self->private->cellrenderer,
					     cell_width,
					     cell_height);

	   column_text = self->private->list->column_data_func (list_item, 1);
	   g_object_set (self->private->cellrenderer, "text", column_text, NULL);
	   g_free (column_text);

	   g_object_set (self->private->cellrenderer, "xalign", 1.0, NULL);
	   g_object_set (self->private->cellrenderer, "yalign", 0.5, NULL);
	   g_object_set (self->private->cellrenderer, "xpad", 0, NULL);
	   g_object_set (self->private->cellrenderer, "ypad", 0, NULL);
	   g_object_set (self->private->cellrenderer, "ellipsize", PANGO_ELLIPSIZE_NONE, NULL);

	   rect[RECT_BG]->x = cell_x;
	   rect[RECT_BG]->y = cell_y;
	   rect[RECT_BG]->width = cell_width;
	   rect[RECT_BG]->height = cell_height;

	   memcpy(rect[RECT_CELL], rect[RECT_BG], sizeof(GdkRectangle));
	   memcpy(rect[RECT_EXPOSE], rect[RECT_BG], sizeof(GdkRectangle));

	   rect[RECT_CELL]->width -= 4; 

	   g_object_set (self->private->cellrenderer, "background-set", FALSE, NULL);

	   gtk_cell_renderer_render (self->private->cellrenderer,
				     self->private->window,
				     self->private->parent,
				     rect[RECT_BG],
				     rect[RECT_CELL],
				     rect[RECT_EXPOSE], 0);
    }

    if ((selected) && (! self->private->use_numbers))
      {

	    draw_rounded_rectangle_pattern (cr,
			0,
			y+1,
			g_object_get_int(self,"width"),
			self->private->row_height-2,
			cr_pattern,
			1.0,
			TRUE,
			4.5);

	    draw_rounded_rectangle (cr,
			0,
			y+1,
			g_object_get_int(self,"width"),
			self->private->row_height-2,
			fg_current,
			0.5,
			1.0,
			FALSE,
			4.5);

     }
   else if (selected)
     {
	    gint sb_width, sb_x;

	    sb_x =
		left_padding + (self->private->font_information->font_width_approx_digits*2);
	    sb_width =
		g_object_get_int(self,"width") - sb_x;

	    if ((sb_x + sb_width) < g_object_get_int(self,"width"))
	      {
		sb_width += g_object_get_int(self,"width") - (sb_x + sb_width);
	      }

	    if ((sb_x + sb_width) > g_object_get_int(self,"width"))
	      {
		sb_width -= (sb_x + sb_width) - g_object_get_int(self,"width");
	      }

	    draw_rounded_rectangle_pattern (cr,
			sb_x - self->private->row_height/4.,
			y+1,
			sb_width + self->private->row_height/4.,
			self->private->row_height-2,
			cr_pattern,
			1.0,
			TRUE,
			4.5);

	    draw_rounded_rectangle (cr,
			sb_x - self->private->row_height/4.,
			y+1,
			sb_width + self->private->row_height/4.,
			self->private->row_height-2,
			fg_current,
			0.5,
			1.0,
			FALSE,
			4.5);
    }

    cairo_pattern_destroy(cr_pattern);
    cairo_destroy (cr);

    for (n=0;n<N_RECTS;n++)
      {
	g_free(rect[n]);
      }
}

static void
blank_window (ChromaList *self)
{
	GdkGC *gc;
	gint w,h;

	g_return_if_fail (self->private->window != NULL);

	/* FIXME: Use Cairo */
	gc =
	    gdk_gc_new (self->private->window);

	gdk_gc_set_rgb_fg_color (gc, bmp_colors_get_color (BMP_SKINCOLOR_PL_BG_NORMAL));
	gdk_drawable_get_size (self->private->window, &w, &h);
	gdk_draw_rectangle (self->private->window,
			gc,
			TRUE,
			0, 0,
			w,
			h);
	gdk_gc_destroy(gc);
}

static void
mark_row_dirty (ChromaList * self, gpointer list_item)
{
    self->private->list->set_dirty_func (list_item, TRUE);
}

static void
mark_row_rendered (ChromaList * self, gpointer list_item)
{
    self->private->list->set_rendered_func (list_item, TRUE);
}

static void
process_cells (ChromaList *self)
{
    g_return_if_fail (self->private->window != NULL);

    if (! self->private->list) {
	blank_window (self);
	return;
    }

    if (self->private->list->items)
      {

	gchar *aux0;
	gint k, n, left_padding, start, end;

	n =
	    g_list_length(self->private->list->items);
        aux0 =
	    g_strdup_printf("%d", n);

	if (self->private->use_numbers)
          {
	    left_padding = 
		  self->private->font_information->font_width_approx_digits * (gint_count_digits (g_list_length(self->private->list->items))) + 8;
	  }
	else
          {
	    left_padding = 0;
	  }

	g_free (aux0);

	start =
	    self->private->list_offset;
	end =
	    self->private->list_offset + chroma_list_get_n_item_display (self);

	if (end > g_list_length (self->private->list->items))
	    end = g_list_length (self->private->list->items);

	for (k = 0; k < (end-start); k++)
          {
            gpointer item;

            item = g_list_nth(self->private->list->items, k+start)->data;
            if (self->private->list->get_dirty_func (item))
	      {
                    render_row (self, k, k+start, left_padding);
		    self->private->list->set_dirty_func (item, FALSE);
              }
	  }

      }
}

static void
mark_visible_dirty_nonrendered (ChromaList *self)
{
    GList *iter_list;
    gpointer item;
    gint n, start, end;

    if (!self->private->list) return;
    if (!self->private->list->items) return;

    start = self->private->list_offset;
    end = start + (self->private->allocation.height/self->private->row_height);

    if (end > g_list_length (self->private->list->items)) end = g_list_length (self->private->list->items);

    for (n=start; n<end;n++){
	item = g_list_nth (self->private->list->items, n)->data;
        if (item)
		mark_row_dirty (self, item);
    }

    iter_list = self->private->list->items;

    while (iter_list) {
	item = iter_list->data;
	mark_row_rendered (self, item);

	iter_list = g_list_next (iter_list);
    }

}

static void
mark_visible_dirty (ChromaList *self)
{
    gpointer item;
    gint n, start, end;

    if (!self->private->list) return;
    if (!self->private->list->items) return;

    start = self->private->list_offset;
    end = start + (self->private->allocation.height/self->private->row_height);

    if (end > g_list_length (self->private->list->items)) end = g_list_length (self->private->list->items);

    for (n=start; n<end;n++){
	item = g_list_nth (self->private->list->items, n)->data;
        if ((item) && (!self->private->list->get_rendered_func(item)))
	{
	    mark_row_dirty (self, item);
	    mark_row_rendered (self, item);
	}
    }
}

static void
recreate_window(ChromaList * self)
{
    G_LOCK(LIST_WINDOW);

    if (self->private->window)
        gdk_drawable_unref(self->private->window);


    if ((! self->private->list) || (! self->private->list->items)) {

	if (GTK_WIDGET_VISIBLE(self->private->parent)) {

	    if (self->private->allocation.width != -1) {

		self->private->window = gdk_pixmap_new(self->private->parent->window,
					   self->private->allocation.width,
					   self->private->allocation.height,
					   gdk_drawable_get_depth (self->private->parent->window));
	    }
	}
	else
	self->private->window = NULL;

    } else {

	if (GTK_WIDGET_VISIBLE(self->private->parent) && (self->private->allocation.width >= 0)) {

           self->private->window = gdk_pixmap_new(self->private->parent->window,
					   self->private->allocation.width,
					   self->private->row_height * chroma_list_get_n_item_display(self),
					   gdk_drawable_get_depth (self->private->parent->window));

	    if ((self->private->list) && (self->private->list->items))
		    mark_visible_dirty (self);
	}
	else
	self->private->window = NULL;
    }

    if (self->private->window) {
	blank_window (self);
        process_cells (self);
    }

    G_UNLOCK(LIST_WINDOW);
}

void
chroma_list_move_resize(ChromaList * self, gint pos_x, gint pos_y, gint w,
                        gint h)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->allocation.x = pos_x;
    self->private->allocation.y = pos_y;
    self->private->allocation.width = w;
    self->private->allocation.height = h;

    recreate_window(self);

    mark_visible_dirty_nonrendered (self);

    chroma_list_emit_expose_request(self);
}

GObject *
chroma_list_new(GtkWidget *parent, gint pos_x, gint pos_y, gint width,
                gint height)
{
    ChromaList *self;
    self =
        g_object_new(chroma_list_get_type(), NULL);

    self->private->parent = parent;

    return G_OBJECT(self);
}

gboolean
chroma_list_inside(ChromaList * self, gint x, gint y)
{
    return ((x >= self->private->allocation.x) &&
            (x <=
             (self->private->allocation.x + self->private->allocation.width))
            && (y >= self->private->allocation.y) &&
            (y <=
             (self->private->allocation.y +
              self->private->allocation.height)));
}

void
chroma_list_emit_changed(ChromaList * self)
{
    g_signal_emit(G_OBJECT(self),
                      chroma_list_signals[CHROMA_LIST_SIGNAL_CHANGED], 0);
    chroma_list_emit_expose_request (self);
}

void
chroma_list_mark_dirty (ChromaList * self)
{
    GList *iter_list;

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (! self->private->list)
	return;

    if (! self->private->list->items)
	return;

    iter_list = self->private->list->items;
    while (iter_list) {
	    self->private->list->set_dirty_func (iter_list->data, TRUE);
	    self->private->list->set_rendered_func (iter_list->data, FALSE);

	    iter_list = g_list_next (iter_list);
    }

    recreate_window (self);
}

void
chroma_list_list_changed (ChromaList * self)
{
    GList *iter_list;

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (! self->private->list)
	return;

    if (! self->private->list->items)
	return;

    iter_list = self->private->list->items;
    while (iter_list) {
	    self->private->list->set_dirty_func (iter_list->data, TRUE);
	    self->private->list->set_rendered_func (iter_list->data, FALSE);
            iter_list = g_list_next(iter_list);
    }

    iter_list = self->private->list->items;

    if ((iter_list) && (g_list_length(iter_list)>0))
      {
	if (g_list_length(iter_list) < chroma_list_get_n_item_display (self)) {
	    g_object_set (G_OBJECT(g_object_get_data(G_OBJECT(self), "adjustment")), "position", 0.0, NULL);
	    self->private->list_offset = 0;
	} else if ((g_list_length (iter_list) - self->private->list_offset) < chroma_list_get_n_item_display(self)) {
	    g_object_set (G_OBJECT(g_object_get_data(G_OBJECT(self), "adjustment")), "position", 1.0, NULL);
	    self->private->list_offset = chroma_list_get_n_items(self) - chroma_list_get_n_item_display(self);
	} else if (self->private->list_offset > g_list_length(iter_list)) {
	    g_object_set (G_OBJECT(g_object_get_data(G_OBJECT(self), "adjustment")), "position", 0.0, NULL);
	    self->private->list_offset = 0;
	} else {
	    gdouble a, b, position;

	    a = (chroma_list_get_n_items(self)-1) - chroma_list_get_n_item_display(self);
	    b = self->private->list_offset;

	    if ( (a>0) && (b>0)) {
		position = b/a;
		g_object_set (G_OBJECT(g_object_get_data(G_OBJECT(self), "adjustment")), "position", position, NULL);
	    }
	}
     }

  if (!iter_list || (iter_list && (g_list_length(iter_list)==0)))
     {
	g_object_set (G_OBJECT(g_object_get_data(G_OBJECT(self), "adjustment")), "position", 0.0, NULL);
     }

    recreate_window (self);
}

static void
unselect_all(gpointer list_item, ChromaList * self)
{
    self->private->list->row_set_selected_func(list_item, FALSE);
}

static void
select_all(gpointer list_item, ChromaList * self)
{
    self->private->list->row_set_selected_func(list_item, TRUE);
}

static void
invert_select(gpointer list_item, ChromaList * self)
{
    gboolean selected;

    selected = self->private->list->row_get_selected_func(list_item);
    self->private->list->row_set_selected_func(list_item,
                                               selected ? FALSE : TRUE);
}

void
chroma_list_invert_selection(ChromaList * self)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    if (!self->private->list)
        return;

    g_list_foreach(self->private->list->items, (GFunc) invert_select, self);
    chroma_list_emit_changed(self);
}

void
chroma_list_select_nth(ChromaList * self, gint _idx, gboolean selected)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->list->
        row_set_selected_func(g_list_nth(self->private->list->items, _idx),
                              selected);

    self->private->list->last_selected = selected ? _idx : -1;
}

void
chroma_list_select_all(ChromaList * self)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    if (!self->private->list)
        return;

    g_list_foreach(self->private->list->items, (GFunc) select_all, self);
    self->private->list->multiple_selected = TRUE;

    chroma_list_emit_expose_request(self);
}

void
chroma_list_unselect_all(ChromaList * self)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    if (!self->private->list)
        return;

    g_list_foreach(self->private->list->items, (GFunc) unselect_all, self);
    self->private->list->last_selected = LAST_SELECTED_NONE;
    self->private->list->multiple_selected = FALSE;
#if 0
    chroma_list_emit_changed(self);
#endif
    chroma_list_emit_expose_request(self);
}

gchar *
chroma_list_get_selected_name(ChromaList * self)
{
    gchar *name;

    name =
        self->private->list->
        column_data_func((gpointer)g_list_nth
                         (self->private->list->items,
                          self->private->list->last_selected), 0);

    return name;
}

static void
select_name(gpointer list_item, SelectNameTuple * tuple)
{
    gchar *name;

    name = tuple->self->private->list->column_data_func(list_item, 0);

    if (!strcmp(name, tuple->name)) {
        tuple->self->private->list->row_set_selected_func(list_item, TRUE);
        tuple->self->private->list->last_selected = tuple->idx;
    }
    tuple->idx++;
}

void
chroma_list_select_name(ChromaList * self, const gchar * name)
{
    SelectNameTuple *tuple;

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (!name)
        return;
    if (!self->private->list)
        return;

    tuple = g_new0(SelectNameTuple, 1);
    tuple->self = self;
    tuple->name = g_strdup(name);
    tuple->idx = 0;

    chroma_list_unselect_all(self);

    g_list_foreach(self->private->list->items, (GFunc) select_name, tuple);
    self->private->list->multiple_selected = FALSE;

    chroma_list_emit_changed(self);

    g_free(tuple);
}

gboolean
chroma_list_dispatch_button_release(ChromaList * self, GdkEventButton * event)
{
    if (event->type != GDK_BUTTON_PRESS)
        return FALSE;

    if (!chroma_list_inside (self, event->x, event->y))
        return FALSE;

    self->private->active = FALSE;
    self->private->grabbed_position = -1;

    return TRUE;
}

gboolean
chroma_list_dispatch_drag_leave(ChromaList * self)
{
    self->private->drag = FALSE;

    return TRUE;
}

gint
chroma_list_get_drop_position(ChromaList * self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    return self->private->drag_row + self->private->list_offset;
}

gboolean
chroma_list_dispatch_drag_motion(ChromaList * self, gint x, gint y)
{
    gint control_y;
    gint position;

    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (!chroma_list_inside(self, x, y))
        return FALSE;

    self->private->drag = TRUE;

    control_y = y - self->private->allocation.y;

    if (self->private->list == NULL)
        return FALSE;

    position = control_y / self->private->row_height;
/*  position += self->private->list_offset; */

    if (position < 0)
        position = 0;
    if (position > chroma_list_get_n_item_display(self))
        position = chroma_list_get_n_item_display(self);
    if (position > g_list_length(self->private->list->items))
        position = g_list_length(self->private->list->items);

    self->private->drag_row = position;

    return TRUE;
}

gboolean
chroma_list_dispatch_key_press(ChromaList *self, GdkEventKey *event)
{
    gpointer list_item;

    g_return_val_if_fail(CHROMA_IS_LIST(self), FALSE);

    if (!self->private->list)
	return FALSE;

    if (!self->private->list->items)
	return FALSE;

    if (self->private->list->last_selected == LAST_SELECTED_NONE)
    {
	self->private->list->last_selected = 0;
	list_item = g_list_nth(self->private->list->items, 0)->data;
	self->private->list->row_set_selected_func(list_item, TRUE);

	chroma_list_emit_cursor_position (self);
	chroma_list_emit_expose_request (self);
	return TRUE;
    }


    switch (event->keyval)
    {
	    case GDK_Return:
	    case GDK_KP_Enter:
	    {
		    if (! self->private->list->multiple_selected)
	            g_signal_emit(self,
                      chroma_list_signals[CHROMA_LIST_SIGNAL_ROW_ACTIVATED],
                      0, self->private->list->last_selected, NULL);
	    }
	    break;

	    case GDK_Up:
	    {
		if (self->private->list->last_selected > 0)
		{
			if (! ((event->state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK))
			{
				if (self->private->list->multiple_selected)
				    g_list_foreach (self->private->list->items, (GFunc)unselect_all, NULL);

				list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
			        self->private->list->row_set_selected_func(list_item, FALSE);
			}
    			else
			{
				gint prev;

				prev = self->private->list->last_selected - 1;

				if (prev >= 0)
				{
					list_item = (g_list_nth(self->private->list->items, prev)->data);
					if (self->private->list->row_get_selected_func(list_item))
					{
						list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	            		    	        self->private->list->row_set_selected_func(list_item, FALSE);
					}


				}
			}

			self->private->list->last_selected--;
	    		list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	    	        self->private->list->row_set_selected_func(list_item, TRUE);

			if (self->private->list->last_selected < self->private->list_offset)
			    scroll_up (self);
			else
			    chroma_list_emit_expose_request (self);

			chroma_list_emit_cursor_position (self);
		}
	    }
	    break;

	    case GDK_Down:
	    {
		if (self->private->list->last_selected < g_list_length(self->private->list->items)-1)
		{
			if (! ((event->state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK))
			{
				if (self->private->list->multiple_selected)
				    g_list_foreach (self->private->list->items, (GFunc)unselect_all, NULL);

				list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
		    	        self->private->list->row_set_selected_func(list_item, FALSE);
			}
			else
			{
				gint next;

				next = self->private->list->last_selected + 1;

				if ((next <= (g_list_length(self->private->list->items)-1)))
				{
					list_item = (g_list_nth(self->private->list->items, next)->data);
					if (self->private->list->row_get_selected_func(list_item))
					{
						list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	            		    	        self->private->list->row_set_selected_func(list_item, FALSE);
					}


				}
			}

			self->private->list->last_selected++;
	    		list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	    	        self->private->list->row_set_selected_func(list_item, TRUE);

			if (self->private->list->last_selected > ((self->private->allocation.height/self->private->row_height)-1))
			    scroll_down (self);
			else
			    chroma_list_emit_expose_request (self);

			chroma_list_emit_cursor_position (self);
		}
	    }
	    break;

	    case GDK_Page_Up:
	    {
		if (self->private->list->last_selected > 0)
		{
#if 0
			if (! ((event->state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK))
			{
				if (self->private->list->multiple_selected)
				    g_list_foreach (self->private->list->items, (GFunc)unselect_all, NULL);

				list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
			        self->private->list->row_set_selected_func(list_item, FALSE);
			}
    			else
			{
				gint prev;

				prev = self->private->list->last_selected - 1;

				if (prev >= 0)
				{
					list_item = (g_list_nth(self->private->list->items, prev)->data);
					if (self->private->list->row_get_selected_func(list_item))
					{
#endif
						list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	            		    	        self->private->list->row_set_selected_func(list_item, FALSE);
#if 0
					}


				}
			}
#endif

			self->private->list->last_selected -= chroma_list_get_n_item_display(self);
			if (self->private->list->last_selected < 0) self->private->list->last_selected = 0;
	    		list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	    	        self->private->list->row_set_selected_func(list_item, TRUE);

			if (self->private->list->last_selected < self->private->list_offset)
			    page_up (self);
			else
			    chroma_list_emit_expose_request (self);

			chroma_list_emit_cursor_position (self);
		}

	    }
	    break;

	    case GDK_Page_Down:
	    {
		if (self->private->list->last_selected < g_list_length(self->private->list->items)-1)
		{
#if 0
			if (! ((event->state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK))
			{
				if (self->private->list->multiple_selected)
				    g_list_foreach (self->private->list->items, (GFunc)unselect_all, NULL);

				list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
		    	        self->private->list->row_set_selected_func(list_item, FALSE);
			}
			else
			{
				gint next;

				next = self->private->list->last_selected + 1;

				if ((next <= (g_list_length(self->private->list->items)-1)))
				{
					list_item = (g_list_nth(self->private->list->items, next)->data);
					if (self->private->list->row_get_selected_func(list_item))
					{
#endif
						list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	            		    	        self->private->list->row_set_selected_func(list_item, FALSE);
#if 0
					}


				}
			}
#endif

			self->private->list->last_selected += chroma_list_get_n_item_display(self);
			if (self->private->list->last_selected > (g_list_length(self->private->list->items)-1)) self->private->list->last_selected = g_list_length (self->private->list->items)-1;
	    		list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	    	        self->private->list->row_set_selected_func(list_item, TRUE);

			if (self->private->list->last_selected > ((self->private->allocation.height/self->private->row_height)-1))
			    page_down (self);
			else
			    chroma_list_emit_expose_request (self);

			chroma_list_emit_cursor_position (self);
		}

	    }
	    break;

	    default:
		break;
    }

    return TRUE;
}


gboolean
chroma_list_dispatch_button_press(ChromaList * self, GdkEventButton * event)
{
    GList *itemarray;
    gpointer list_item;
    gint control_y;
    gint position;

    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (!chroma_list_inside (self, event->x, event->y)) 
      {
        self->private->active = FALSE;
        return FALSE;
      }

    self->private->active = TRUE;

    control_y = event->y - self->private->allocation.y;

    if (self->private->list == NULL)
        return FALSE;

    position = control_y / self->private->row_height;
    position += self->private->list_offset;

    itemarray = self->private->list->items;

    if (position < 0)
        return FALSE;

    if (position > (g_list_length(self->private->list->items) - 1) &&
        (event->type == GDK_BUTTON_PRESS) && (self->private->selection_mode == GTK_SELECTION_SINGLE)) {

	if ((self->private->list->last_selected >= 0) && (self->private->list->last_selected <= (g_list_length(self->private->list->items)-1))) {

            list_item = (g_list_nth(self->private->list->items, self->private->list->last_selected)->data);
	    self->private->list->row_set_selected_func(list_item, FALSE);
	}

        position =
	    g_list_length(self->private->list->items) - 1;
        list_item =
	    (g_list_nth(self->private->list->items, position)->data);

        self->private->list->row_set_selected_func(list_item, TRUE);
        self->private->list->last_selected = position;

	if (self->private->change_on_select)
	    chroma_list_emit_changed(self);
	else
	    chroma_list_emit_expose_request(self);

	chroma_list_emit_cursor_position (self);
        return TRUE;
    }
    else if (position > (g_list_length(self->private->list->items) - 1) &&
             (event->type == GDK_2BUTTON_PRESS)) {

	g_list_foreach(self->private->list->items, (GFunc)unselect_all, self);

        position =
	    g_list_length(self->private->list->items) - 1;
        list_item =
	    (g_list_nth(self->private->list->items, position)->data);

        self->private->list->row_set_selected_func(list_item, TRUE);
        self->private->list->last_selected = position;

	if (self->private->change_on_select)
	    chroma_list_emit_changed(self);
	else
	    chroma_list_emit_expose_request(self);

	chroma_list_emit_cursor_position (self);

        g_signal_emit(self,
                      chroma_list_signals[CHROMA_LIST_SIGNAL_ROW_ACTIVATED],
                      0, position, NULL);
        return TRUE;
    }

    if (!g_list_nth(itemarray, position))
        return FALSE;

    self->private->grabbed_position = position;

    if (event->type == GDK_2BUTTON_PRESS) {

	if (self->private->change_on_select)
	    chroma_list_emit_changed(self);
	else
	    chroma_list_emit_expose_request(self);

        g_signal_emit(self,
                      chroma_list_signals[CHROMA_LIST_SIGNAL_ROW_ACTIVATED],
                      0, position, NULL);
        return TRUE;
    }

    if (event->type != GDK_BUTTON_PRESS)
        return FALSE;

    /* Simple select */
    if (((event->state & GDK_SHIFT_MASK) == 0) &&
        ((event->state & GDK_CONTROL_MASK) == 0) && (event->button == 1)) {
        if ((self->private->list->last_selected != LAST_SELECTED_NONE) &&
            ((self->private->list->multiple_selected) || (self->private->selection_mode == GTK_SELECTION_SINGLE))) {
            g_list_foreach(self->private->list->items, (GFunc)unselect_all, self);
        }
        else if (self->private->list->last_selected >= 0) {
	    if (!self->private->list) return FALSE;
	    if (!self->private->list->items) return FALSE;
	    if (self->private->list->last_selected > (g_list_length (self->private->list->items)-1)) return FALSE;
            list_item =
                (g_list_nth
                 (self->private->list->items,
                  self->private->list->last_selected)->data);
            if (list_item)
	    {
                self->private->list->row_set_selected_func(list_item, FALSE);
	    }
        }

        list_item = (g_list_nth(self->private->list->items, position)->data);
        if (!list_item)
            return FALSE;

        if (self->private->list->row_get_selected_func(list_item) == FALSE) {
            self->private->list->row_set_selected_func(list_item, TRUE);
            self->private->list->last_selected = position;
        }
        else {
            self->private->list->last_selected = LAST_SELECTED_NONE;
        }

        self->private->list->multiple_selected = FALSE;

	if (self->private->change_on_select)
	    chroma_list_emit_changed(self);
	else
	    chroma_list_emit_expose_request(self);

	chroma_list_emit_cursor_position (self);

        return TRUE;
    }

    /* Select multiple using CTRL */
    if (((event->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK) &&
        (event->button == 1)) {
        gint position;
        gpointer list_item;
        gboolean selected;

        position = control_y / self->private->row_height;
        position += self->private->list_offset;

        if ((self->private->selection_mode == GTK_SELECTION_SINGLE) &&
            (position != self->private->list->last_selected))
            return FALSE;

        list_item = (g_list_nth(self->private->list->items, position)->data);
        selected = self->private->list->row_get_selected_func(list_item);
        self->private->list->row_set_selected_func(list_item, !selected);

        if (self->private->list->last_selected == position)
            self->private->list->multiple_selected = FALSE;
        else
            self->private->list->multiple_selected = TRUE;

	if (self->private->change_on_select)
	    chroma_list_emit_changed(self);
	else
	    chroma_list_emit_expose_request(self);
        return TRUE;
    }

    if (self->private->selection_mode == GTK_SELECTION_SINGLE)
        return FALSE;

    /* Select multiple using SHIFT */
    if (((event->state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK) &&
        (event->button == 1)) {
        gpointer *list_item;
        gint position;

        if (self->private->list->last_selected != LAST_SELECTED_NONE) {
            GList *itemarray;

            position = control_y / self->private->row_height;
            position += self->private->list_offset;
            itemarray = self->private->list->items;

            if (self->private->list->last_selected > position) {
                gint n;
                for (n = self->private->list->last_selected; n >= position; n--) {
                    list_item = (g_list_nth(itemarray, n)->data);
                    self->private->list->row_set_selected_func(list_item,
                                                               TRUE);
                }
            }
            else if (self->private->list->last_selected < position) {
                gint n;
                for (n = self->private->list->last_selected; n <= position; n++) {
                    list_item = (g_list_nth(itemarray, n)->data);
                    self->private->list->row_set_selected_func(list_item,
                                                               TRUE);
                }

            }
        }

        self->private->list->multiple_selected = TRUE;

	if (self->private->change_on_select)
	    chroma_list_emit_changed(self);
	else
	    chroma_list_emit_expose_request(self);
        return TRUE;
    }

    chroma_list_emit_expose_request(self);
    return TRUE;
}

GList *
chroma_list_get_selected_paths(ChromaList * self)
{
    GList *paths = NULL;
    gpointer list_item;
    gboolean selected = FALSE;
    gint n;

    g_return_val_if_fail(CHROMA_IS_LIST(self), NULL);

    if (!self->private->list)
        return NULL;

    if (!self->private->list->items)
        return NULL;

    for (n = 0; n < g_list_length(self->private->list->items); n++) {
        GtkTreePath *tree_path;

        /* FIXME: Hmm... */
        list_item = (g_list_nth(self->private->list->items, n)->data);
        selected = self->private->list->row_get_selected_func(list_item);

        if (selected) {
            tree_path = gtk_tree_path_new();
            gtk_tree_path_append_index(tree_path, n);
            paths = g_list_append(paths, tree_path);
        }
    }

    return paths;
}

gboolean
chroma_list_dispatch_motion(ChromaList * self, GdkEventMotion * event)
{
    gint control_y, position;

    g_return_val_if_fail(CHROMA_IS_LIST(self), FALSE);

    if (self->private->selection_mode == GTK_SELECTION_SINGLE) return FALSE;

    if (self->private->list == NULL)
        return FALSE;

    if (!chroma_list_inside (self, event->x, event->y))
        return FALSE;

    if (!self->private->active)
        return FALSE;

    control_y = event->y - self->private->allocation.y;
    position = control_y / self->private->row_height;
    position += self->private->list_offset;

    if (self->private->grabbed_position == position)
        return FALSE;

    if (position < 0)
        position = 0;
    if (position > (g_list_length(self->private->list->items) - 1))
        position = g_list_length(self->private->list->items) - 1;

    if (abs(self->private->grabbed_position - position)>1)
    {
	    if (position < self->private->grabbed_position)
		position = self->private->grabbed_position-1;
	    else
		position = self->private->grabbed_position+1;
    }

    self->private->list->item_swap_func(self->private->list->items,
                                        self->private->grabbed_position,
                                        position);

    if (self->private->list->current == self->private->grabbed_position)
        self->private->list->current = position;
    else if (self->private->list->current == position)
        self->private->list->current = self->private->grabbed_position;

    bmp_system_control_rows_swapped (bmp_system_control,
				     g_object_get_int (bmp_system_control, "current_tracklist"),
				     self->private->grabbed_position,
				     position);

    self->private->grabbed_position = position;
    self->private->list->last_selected = position;

    chroma_list_emit_expose_request(self);
    return TRUE;
}

void
chroma_list_draw(ChromaList * self, GtkWidget * widget, cairo_t * cr,
                 GdkVisibilityState state)
{
    gint x1_line = 0, y1_line = 0,
         x2_line = 0, y2_line = 0,
         left_padding;
    double r, g, b, a;

    g_return_if_fail(CHROMA_IS_LIST(self));

    if (! self->private->list)
	return;

    if (! self->private->list->items) {
	blank_window (self);
    }
    else {
	process_cells (self);
    }

    if (self->private->window) {

        gdk_draw_drawable(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
                       self->private->window, 0, 0,
                       g_object_get_int(self, "pos_x"),
		       g_object_get_int(self, "pos_y"),
                       -1,
		       -1);
    }

    if (self->private->use_numbers) {
	left_padding = 
	    self->private->font_information->font_width_approx_digits * (gint_count_digits (g_list_length(self->private->list->items))) + 8;
        x1_line =
	    left_padding + (1.5*self->private->font_information->font_width_approx_digits) + 8; 

	if (pango_font_description_get_style(self->private->font_information->font_description) != PANGO_STYLE_NORMAL)
	    x1_line += self->private->font_information->font_width_approx_digits;

        y1_line = g_object_get_int (self,"pos_y");
        x2_line = x1_line;
        y2_line = y1_line + self->private->allocation.height;

	{
	    cairo_save (cr);
	    cairo_set_antialias (cr, CAIRO_ANTIALIAS_NONE);
	    cairo_set_line_width(cr, 1.0);
	    cairo_move_to (cr, x1_line, y1_line);
	    cairo_line_to (cr, x1_line, y2_line);
	    gdk_color_to_rgba (bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT), &r, &g, &b, &a);
	    cairo_set_source_rgba (cr, r,g,b,0.6);
	    cairo_stroke (cr);
	    cairo_restore (cr);
	}

    }
    else
	left_padding = 0;

    gdk_color_to_rgba(bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT), &r, &g, &b, &a);

    if ((self->private->drag) &&
        (self->private->drop_mode == CHROMA_LIST_DROP_ROWS)) {
        gint lx, ly, lw;

	lx = x1_line + self->private->font_information->font_width_approx_digits;
        ly = g_object_get_int(self,
                              "pos_y") +
            (self->private->row_height * self->private->drag_row) + 1;

	lw = g_object_get_int(self,"width") - x1_line + self->private->font_information->font_width_approx_digits;

        cairo_set_line_width(cr, 1);
        cairo_move_to(cr, lx, ly);
        cairo_line_to(cr, lx + lw, ly);
        cairo_set_source_rgba(cr, r, g, b, 1.0);
	cairo_set_line_cap (cr, CAIRO_LINE_CAP_ROUND);
	cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
        cairo_stroke(cr);
    }
    else if ((self->private->drag) &&
             (self->private->drop_mode == CHROMA_LIST_DROP_WHOLE)) {
        gint lx, ly, lw, lh;
        lx = g_object_get_int(self, "pos_x")+1;
        ly = g_object_get_int(self, "pos_y")+1;
        lw = g_object_get_int(self, "width")-2;
        lh = g_object_get_int(self, "height")-2;
        draw_rounded_rectangle(cr, lx, ly, lw, lh, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT), 1.0, 1.5, FALSE, 12);
    }

}

void
chroma_list_set_font_information(ChromaList * self,
                                 BmpFontInformation * font_information)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    /* if (self->private->font_information != NULL) g_free (self->private->font_information); */

    self->private->font_information = font_information;

    self->private->row_height =
        self->private->font_information->font_height + 3; 

    g_signal_emit(self,
                  chroma_list_signals[CHROMA_LIST_SIGNAL_EXPOSE_REQUEST], 0,
                  g_object_get_int(self, "pos_x"), g_object_get_int(self,
                                                                    "pos_y"),
                  g_object_get_int(self, "width"), g_object_get_int(self,
                                                                    "height")
        );
}

void
chroma_list_set_use_numbers(ChromaList * self, gboolean use_numbers)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->use_numbers = use_numbers;

    chroma_list_mark_dirty (self);

    g_signal_emit(self,
                  chroma_list_signals[CHROMA_LIST_SIGNAL_EXPOSE_REQUEST], 0,
                  g_object_get_int(self, "pos_x"), g_object_get_int(self,
                                                                    "pos_y"),
                  g_object_get_int(self, "width"), g_object_get_int(self,
                                                                    "height"));
}

gint
chroma_list_get_n_item_display(ChromaList * self)
{
    gint n_items;

    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (! self->private->row_height) return 0;

    n_items = g_object_get_int(self, "height") / self->private->row_height;

    return n_items;
}

gint
chroma_list_get_n_items(ChromaList * self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (!self->private->list)
        return 0;
    if (!self->private->list->items)
        return 0;

    return g_list_length(self->private->list->items);
}

const BmpListType *
chroma_list_get_list(ChromaList * self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), NULL);

    return self->private->list;
}

void
chroma_list_set_list(ChromaList * self, const BmpListType * list)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->list = BMP_LIST(list);
    g_object_set (self, "list_offset", (gint)0, NULL);

    if (list) {
        g_object_set (g_object_get_data(G_OBJECT(self), "adjustment"), "position", list->adjustment, NULL);
    }

    recreate_window (self);
    mark_visible_dirty_nonrendered (self);
    chroma_list_list_changed (self);
    chroma_list_emit_expose_request(self);
}

gint
chroma_list_get_last_selected(ChromaList * self)
{
    g_return_val_if_fail(CHROMA_IS_LIST(self), -1);

    if (self->private->list->multiple_selected)
        return -1;
    return self->private->list->last_selected;
}

void
chroma_list_set_last_selected(ChromaList * self, gint _idx)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->list->last_selected = _idx;
}

void
chroma_list_set_drop_mode(ChromaList * self, ChromaListDropMode mode)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->drop_mode = mode;
}

void
chroma_list_set_change_on_select (ChromaList *self, gboolean change_on_select)
{
    g_return_if_fail(CHROMA_IS_LIST(self));

    self->private->change_on_select = change_on_select;
}

gint
chroma_list_get_row_height (ChromaList *self)
{
    if (! CHROMA_IS_LIST(self))
	return 0;

    return self->private->row_height;
}

void
chroma_list_get_size (ChromaList *self, GtkAllocation *allocation)
{
    if (! CHROMA_IS_LIST(self))
    {
	allocation = NULL;
	return;
    }

    memcpy(allocation, &(self->private->allocation), sizeof(GtkAllocation));
}

static void
vs_move (ChromaVScrollbar *vs, gint total, gint offset)
{
    gdouble position, ceil_val;
    gint r_offset;

    ceil_val = 0.00;

    if (offset <= 0)
	position = 0;
    else
    {
	gboolean done = FALSE;

	/* FIXME: We adapt dynamically here which sucks */
	while (! done)
	{
		position = (offset/total)/100. + ceil_val;

		/* Reverse calculation */
	       r_offset = (total) * position;

	       if (r_offset == offset)
		    done = TRUE;
	       else
		    ceil_val += 0.0001;
	}

    }

    if (position < 0) position = 0;
    if (position > 1) position = 1;
    g_object_set(G_OBJECT(vs),"position", position, NULL);

}

static void
scroll_down (ChromaList *self)
{
    ChromaVScrollbar *vs;
    gint offset, n_items, n_item_display;

    vs = g_object_get_data (G_OBJECT(self), "adjustment");
    n_items = g_object_get_int (self, "n_items");
    n_item_display = chroma_list_get_n_item_display(self);
    offset = self->private->list_offset + 1;

    vs_move (vs, n_items-n_item_display, offset);
}


static void
scroll_up (ChromaList *self)
{
    ChromaVScrollbar *vs;
    gint offset, n_items, n_item_display;

    vs = g_object_get_data (G_OBJECT(self), "adjustment");
    n_items = g_object_get_int (self, "n_items");
    n_item_display = chroma_list_get_n_item_display(self);
    offset = self->private->list_offset - 1;

    vs_move (vs, n_items-n_item_display, offset);
}

static void
page_up (ChromaList *self)
{
    ChromaVScrollbar *vs;
    gint offset, n_items, n_item_display;

    vs = g_object_get_data (G_OBJECT(self), "adjustment");
    n_items = g_object_get_int (self, "n_items");
    n_item_display = chroma_list_get_n_item_display(self);
    offset = self->private->list_offset - n_item_display;

    vs_move (vs, n_items-n_item_display, offset);
}

static void
page_down (ChromaList *self)
{
    ChromaVScrollbar *vs;
    gint offset, n_items, n_item_display;

    vs = g_object_get_data (G_OBJECT(self), "adjustment");
    n_items = g_object_get_int (self, "n_items");
    n_item_display = chroma_list_get_n_item_display(self);
    offset = self->private->list_offset + n_item_display;

    vs_move (vs, n_items-n_item_display, offset);
}

static void
on_vscrollbar_position (ChromaVScrollbar *vs, gdouble position, ChromaList *self)
{
	gint n_items, n_item_display, offset;

	n_items =
	    g_object_get_int (self, "n_items");
	n_item_display =
	    chroma_list_get_n_item_display(self);

	if (n_items < n_item_display) return;

	offset = (n_items - n_item_display) * position;

	if (offset >= 0)
	 	g_object_set (self, "list_offset", offset, NULL);

	self->private->list->adjustment = position;
}

void
chroma_list_set_adjustment (ChromaList *self, ChromaVScrollbar *adjustment)
{
    g_object_set_data (G_OBJECT(self), "adjustment", adjustment);
    g_signal_connect (adjustment, "position", G_CALLBACK(on_vscrollbar_position), self);
}


/* GObject stuff */
static void
chroma_list_set_property (GObject *object,
                          guint property_id,
                          const GValue * value,
                          GParamSpec * pspec)
{
    ChromaList *self = (ChromaList *) object;

    switch (property_id) {

    case C_LIST_PROP_LIST_OFFSET: {

	    gint offset_old, delta;

	    offset_old =
		self->private->list_offset;

	    /* Create the delta between the old and new offset */
	    delta =
		g_value_get_int(value) - offset_old;

	    if (delta == 0)
	      {
		break;
	      }

	    if (!self->private->list)
	      {
		break;
	      }

	    /* There is no area to copy, we have to re-render wholly */
	    if (abs(delta)>chroma_list_get_n_item_display(self)) {
		mark_visible_dirty (self);
                self->private->list_offset = g_value_get_int(value);
		chroma_list_emit_expose_request(self);
		break;
	    }

	    /* If the delta is positive, we need to move a lower part upwards */
	    if (delta > 0) {
		    gint start, height, n, n_start, n_end;
		    gpointer item;

		    start =
			delta;
		    height =
			chroma_list_get_n_item_display(self)-start;

		    gdk_draw_drawable (self->private->window,
				       self->private->parent->style->fg_gc[GTK_STATE_NORMAL],
				       self->private->window,
				       0, start * self->private->row_height,
				       0, 0,
				       self->private->allocation.width, height * self->private->row_height);

		    self->private->list_offset = g_value_get_int(value);

		    n_start =
			self->private->list_offset + chroma_list_get_n_item_display(self) - delta;
		    n_end =
			self->private->list_offset + chroma_list_get_n_item_display(self);

		    for (n = n_start; n < n_end; n++) {
	                item = g_list_nth(self->private->list->items, n);
			if (item)
			    mark_row_dirty (self, ((GList*)item)->data);
		    }

	    } else {
		    gint start, height, n, n_start, n_end;
		    gpointer item;

		    start =
			abs(delta);
		    height =
			chroma_list_get_n_item_display(self)-start;

		    gdk_draw_drawable (self->private->window,
				       self->private->parent->style->fg_gc[GTK_STATE_NORMAL],
				       self->private->window,
				       0, 0,
				       0, start * self->private->row_height,
				       self->private->allocation.width, height * self->private->row_height);

		    self->private->list_offset = g_value_get_int(value);

		    n_start =
			self->private->list_offset;
		    n_end =
			self->private->list_offset + start;

		    for (n = n_start; n < n_end; n++) {
	                item = g_list_nth(self->private->list->items, n);
			if (item)
			    mark_row_dirty (self, ((GList*)item)->data);
		    }
	    }

            chroma_list_emit_expose_request(self);
        }
        break;

    default:
        /* We don't have any other property... */
        g_assert(FALSE);
        break;
    }
}

static void
chroma_list_get_property(GObject * object,
                         guint property_id,
                         GValue * value, GParamSpec * pspec)
{


    switch (property_id) {

    case C_LIST_PROP_N_ITEMS:
        {
            if (CHROMA_LIST(object)->private->list == NULL)
                g_value_set_int(value, (gint) 0);
            else
                g_value_set_int(value,
                                g_list_length(CHROMA_LIST(object)->private->
                                              list->items));
        }
        break;

    case C_LIST_PROP_ROW_HEIGHT:
        {
            g_value_set_int(value, CHROMA_LIST(object)->private->row_height);
        }
        break;

    case C_LIST_PROP_X:
        {
            g_value_set_int(value, CHROMA_LIST(object)->private->allocation.x);
        }
        break;

    case C_LIST_PROP_Y:
        {
            g_value_set_int(value, CHROMA_LIST(object)->private->allocation.y);
        }
        break;

    case C_LIST_PROP_W:
        {
            g_value_set_int(value, CHROMA_LIST(object)->private->allocation.width);
        }
        break;

    case C_LIST_PROP_H:
        {
            g_value_set_int(value, CHROMA_LIST(object)->private->allocation.height);
        }
        break;

    default:
        /* We don't have any other property... */
        g_assert(FALSE);
        break;
    }
}

static void
chroma_list_dispose(GObject * obj)
{
    ChromaList *self = (ChromaList *) obj;

    if (self->private->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
    /* Make sure dispose does not run twice. */
    self->private->dispose_has_run = TRUE;

    /*
     * In dispose, you are supposed to free all types referenced from this
     * object which might themselves hold a reference to self. Generally,
     * the most simple solution is to unref all members on which you own a
     * reference.
     */
}

static void
chroma_list_finalize(GObject * obj)
{
    ChromaList *self = (ChromaList *) obj;

    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */
    g_free(self->private);
}

static GObject *
chroma_list_constructor(GType type,
                        guint n_construct_properties,
                        GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        /* Invoke parent constructor. */
        ChromaListClass *klass;
        GObjectClass *parent_class;
        klass = CHROMA_LIST_CLASS(g_type_class_peek(CHROMA_TYPE_LIST));
        parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));
        obj = parent_class->constructor(type,
                                        n_construct_properties,
                                        construct_properties);
    }

    /* do stuff. */

    return obj;
}

static void
chroma_list_class_init(ChromaListClass *g_class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);

    GParamSpec *pspec;

    gobject_class->set_property = chroma_list_set_property;
    gobject_class->get_property = chroma_list_get_property;
    gobject_class->dispose = chroma_list_dispose;
    gobject_class->finalize = chroma_list_finalize;
    gobject_class->constructor = chroma_list_constructor;

    pspec = g_param_spec_int("pos_x",
                             "pos_x",
                             "pos_x", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class, C_LIST_PROP_X, pspec);

    pspec = g_param_spec_int("pos_y",
                             "pos_y",
                             "pos_y", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class, C_LIST_PROP_Y, pspec);

    pspec = g_param_spec_int("width",
                             "width",
                             "width", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class, C_LIST_PROP_W, pspec);

    pspec = g_param_spec_int("height",
                             "height",
                             "height", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class, C_LIST_PROP_H, pspec);

    pspec = g_param_spec_int("list_offset",
                             "list_offset",
                             "list_offset", 0, G_MAXINT, 0, G_PARAM_WRITABLE);
    g_object_class_install_property(gobject_class,
                                    C_LIST_PROP_LIST_OFFSET, pspec);

    pspec = g_param_spec_int("n_items",
                             "n_items",
                             "n_items", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class,
                                    C_LIST_PROP_N_ITEMS, pspec);

    pspec = g_param_spec_int("row_height",
                             "row_height",
                             "row_height", 0, G_MAXINT, 0, G_PARAM_READABLE);
    g_object_class_install_property(gobject_class,
                                    C_LIST_PROP_ROW_HEIGHT, pspec);


    chroma_list_signals[CHROMA_LIST_SIGNAL_EXPOSE_REQUEST] =
        g_signal_new("expose_request",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, expose_request),
                     NULL, NULL,
                     bmpx_VOID__INT_INT_INT_INT,
                     G_TYPE_NONE, 4, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT,
                     G_TYPE_INT);

    chroma_list_signals[CHROMA_LIST_SIGNAL_CHANGED] =
        g_signal_new("changed",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, changed),
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

    chroma_list_signals[CHROMA_LIST_SIGNAL_CURSOR_POSITION] =
        g_signal_new("cursor-position",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, cursor_position),
                     NULL, NULL,
                     g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);

    chroma_list_signals[CHROMA_LIST_SIGNAL_ROW_ACTIVATED] =
        g_signal_new("row_activated",
                     G_OBJECT_CLASS_TYPE(gobject_class),
                     G_SIGNAL_RUN_FIRST,
                     G_STRUCT_OFFSET(ChromaListClass, row_activated),
                     NULL, NULL,
                     g_cclosure_marshal_VOID__INT,
                     G_TYPE_NONE, 1, G_TYPE_INT);

}


static void
chroma_list_init(ChromaList *self)
{
    self->private = g_new(ChromaListPrivate, 1);
    self->private->dispose_has_run = FALSE;
    self->private->list = NULL;
    self->private->use_numbers = TRUE;
    self->private->font_information = NULL;
    self->private->list_offset = 0;
    self->private->align = CHROMA_LIST_ALIGN_LEFT;
    self->private->columns = 1;
    self->private->selection_mode = GTK_SELECTION_MULTIPLE;
    self->private->change_on_select = FALSE;
    self->private->active = FALSE;
    self->private->grabbed_position = 0;
    self->private->drag = FALSE;
    self->private->drag_row = -1;
    self->private->changed = TRUE;
    self->private->cellrenderer = gtk_cell_renderer_text_new();
    self->private->drop_mode = CHROMA_LIST_DROP_ROWS;
    self->private->parent = NULL;
    self->private->window = NULL;
    self->private->allocation.width = -1;
}
